/* dvitovdu.c - 18:20 GMT +10:00 Wed 18 Aug 1993 - modifier Geoffrey Tobin */

#include <signal.h>
#include <values.h>
#include <time.h>

#include "version.h"

#include "config.h"
#include "defaults.h"
#include "globals.h"
#include "help.h"
#include "screenio.h"
#include "options.h"
#include "vdu.h"
#include "dvireader.h"
#include "fontreader.h"
#include "lstring.h"
#include "new.h"
#include <ggi/ggi.h>

extern ggi_visual_t *GGIvisual;
void InitGGI(void);

/* Variables Exported, via config.h */

int P_argc = 0;
char ** P_argv = (char **) NULL;
Anyptr __MallocTemp__ = (Anyptr) NULL;


/* Variables Exported, via globals.h */

/* Variables Exported, via vdu.h */

int DVIstatusl, windowstatusl, messagel, commandl, bottoml;
int windowh, windowv, windowwd, windowht;


static void DisplayPage (void);
static void NextCommandLine (void);
static boolean PrintFile __((Char * filename));
static boolean PrintText __((Char * * text));

/* list of signal conditions */

static int common_signals[]
  = {SIGABRT, SIGFPE, SIGILL, SIGINT, SIGSEGV, SIGTERM};

/*******************************************************************************
   DECLARATIONS FOR PROCESSING USER COMMANDS

   Most commands consist of one or two characters and can be entered
   in upper or lowercase.  Multiple commands are processed in the order
   given but we only update the window, if necessary, at the end.
   If a bad command is encountered, any further commands are ignored.
   Some commands can have parameters; they are all dimensions in terms of the
   current units.  Spaces before and after commands and parameters are ignored.
*/

/* Possible commands are:                                                  */
/* i                  a positive integer; display ith DVI page             */
#include<ggi/keyboard.h>
#define NextCmd         GIIK_PageDown
#define PrevCmd         GIIK_PageUp
    /* display next DVI page, depending on direction        */
#define Window          "W"
    /* move window's top left corner to given position      */
#define UpCmd           GIIK_Up
#define DownCmd         GIIK_Down
#define LeftCmd         GIIK_Left
#define RightCmd        GIIK_Right
#define ZoomInCmd       '+'
#define ZoomOutCmd      '-'
    /* move window right a given amount                     */
#define Hsize           "H"
    /* set scaledwd: window's horizontal size               */
#define Vsize           "V"
    /* set scaledht: window's vertical size                 */
#define AutoView        "A"
    /* enable/disable automatic view after page selection   */
#define ZoomInOut       "Z"
    /* halve/double window dimensions                       */
#define Terse           "T"
#define C_Terse         'T'
    /* display quick and nasty chars at reference points    */
#define Box             "B"
#define C_Box           'B'
    /* display box outlines of glyphs                       */
#define Full            "F"
#define C_Full          'F'
    /* display all pixels in glyphs                         */
#define Ic              "I"
    /* get/show dimensions in inches                        */
#define Cm              "C"
    /* get/show dimensions in centimetres                   */
#define Mm              "M"
    /* get/show dimensions in millimetres                   */
#define PcPtPx          "P"
    /* get/show dimensions in picas/points/pixels           */
#define Help            "?"
    /* display help on available commands                   */
#define Show            "S"
    /* display useful statistics                            */
#define Quit            "Q"
    /* have a guess                                         */

#define commprompt      "Command: "

static int command;
int argument;
static int maxpix;
int ReadCommand(void);
/* maximum absolute pixel value;
                                    depends on resolution                  */
/* These flags are used to handle multiple commands:                       */
static boolean screenjustcleared;
    /* has screen just been cleared?          */
static boolean paintDVIStatus;   /* does DVI status line need updating?    */
static boolean paintWindowStatus;
    /* does window status line need updating? */
static boolean paintwindow;   /* does window region need updating?      */
static boolean pageoffpaper;   /* is page off paper?                     */
static boolean badcommand;   /* was there a bad command?               */


/*******************************************************************************
   DECLARATIONS FOR DISPLAYING A PAGE

   The reference points of characters and rules on a page are stored as
   pairs of horizontal and vertical paper pixel coordinates.
   The paper coordinate scheme is described in detail in DVIReader.
   The screen coordinate scheme is described in detail in VDU.
   To update the window region, DVItoVDU maps visible paper pixels
   to screen pixels using windowh and windowv to help with translation,
   and windowwd and windowht to help with scaling.
   What the user sees depends on the current displaymode, the current size
   of the window region (scaledwd by scaledht are in paper pixels and determine
   the horizontal and vertical scaling factors), and the current paper position
   of the window region's top left corner; i.e., (windowleft,windowtop).

   A NOTE ON THE SCALING METHOD USED BY DVItoVDU:
   We desire the following conditions when scaling paper pixels to
   screen pixels:
   1. Rules/glyphs having the same top/bottom/left/right paper coordinates also
      have the same screen coordinates (e.g., to ensure baselines line up).
      This condition is incompatible with a rule/glyph staying the same
      width and height as the window position changes!  Too bad.
   2. After being scaled, visible pixel positions must not exceed the
      window region's edges.  In our case, only the bottom and right edges are
      a problem because scaling starts at the top left corner of the window.
      We use two different scaling calculations depending on
      whether the h/vscalefactors are < 1.0 or not.
   3. Scaled heights and widths must be > 0 even when h/vscalefactors
      approach 0.  If h/vscalefactors are > 1.0 then the width/height of
      paper pixels increase accordingly.
*/

/* terse - show quick and nasty chars at ref pts  */
/* box   - show box outlines of glyphs            */
/* full  - show all pixels in glyphs              */

#define  tersemode  1
#define  boxmode    2
#define  fullmode   4
int displaymode;

/* we get/show dimensions in currentunits */
units currentunits;

static int papertop, paperleft, paperbottom, paperright;
    /* these define the edges of the paper    */

static int windowtop, windowleft, windowbottom, windowright;
    /* these define the current window edges  */

static int invalidtop, invalidleft, invalidbottom, invalidright;
    /* these define the region that has to be redrawn */

static boolean allpagevisible;   /* is all of page visible in window?      */
static boolean outsidepage;      /* is entire window outside page?         */
static int scaledht;             /* current window height in paper pixels  */
static int scaledwd;             /* current window width in paper pixels   */
static double vscalefactor;      /* windowht / scaledht                    */
static double hscalefactor;      /* windowwd / scaledwd                    */
static ruleinfo *thisruleinfo;   /* current rule info in rulelist          */
static fontinfo *unusedfont;     /* first unused font in sorted fontlist   */
static fontinfo *thisfontinfo;   /* current font info in sorted fontlist   */
static charinfo *thischarinfo;   /* current char info in charlist          */
static int thischar;             /* current index into current chartable   */
static boolean fontopen;         /* is thisfontinfo->fontspec open?        */
static boolean useraborted;      /* did user abort page display?           */


/******************************************************************************/

static Char Cap (Char ch)
{
  /* If ch is a lowercase letter, then return capital, else return ch. */

  if (islower(ch))
    return (Char) toupper (ch);
  else
    return ch;
}
/* Cap */

/******************************************************************************/

void PadMesg (void)
{
  /* Pad current text line with spaces. */
  string padding;
  int i;
  for (i=0; i < maxstring; i++)
    padding[i] = ' ';
  padding[maxstring] = '\0';
  MesgString (padding);
}
/* Pad Mesg */

/******************************************************************************/

static void InitWinVars (void)
{
  /* TeX will not generate dimensions > than about 38 feet, so we
     choose an absolute limit on our dimensions to be 40 feet.
  */

  int ymaxpix = yres * (40 * 12);  /* 40 ft = 40 * 12 in */
  maxpix = xres * (40 * 12);
  if (ymaxpix < maxpix)
    maxpix = ymaxpix;

  /* top left corner of paper is fixed at (-1",-1") */
  papertop = -yres;
  paperleft = -xres;

  paperbottom = papertop + paperht - 1;
  paperright = paperleft + paperwd - 1;

  /* User sees the following status values before requesting the first page. */

  /*
     FIRST STATUS LINE:  # pages, DVI page, TeX page, Auto, Displaymode, Zoom.
     Note that DVIReader has already counted the number of pages,
     and initialized currDVIpage and currTeXpage.
  */

  displaymode = fullmode;  /* Full mode, neither Box nor Terse mode */

  /*
     SECOND STATUS LINE:  Window loc. & dims, Page loc. & dims, Unit.
  */

  windowleft = 0;   /* window location */
  windowtop = 0;
  scaledwd = windowwd;   /* window size is initially unscaled */
  scaledht = windowht;

  minhp = 0;
  minvp = 0;   /* page location */
  maxhp = 0;
  maxvp = 0;
  hscalefactor = 1.0;   /* page size is initially unscaled, too */
  vscalefactor = 1.0;

  currentunits = ic;    /* inch */

  /* WINDOW:  Chars & Rules. */

  pageempty = true;  /* no page read yet, so no chars and no rules! */
}
/* InitWinVars */

/******************************************************************************/

static void ClearMessageLine (void)
{
  /* Clear message line and move cursor to start of line.
     We don't show any message here; that will usually be done
     immediately after calling this routine.
  */

  ClearTextLine (messagel);
  MoveToTextLine (messagel);
}
/* ClearMessageLine */

/******************************************************************************/

static void WaitForReturn (void)
{
  int ch;
  MesgFlush();
  do {
    ch=ggiGetc(GGIvisual);
  } while (ch != GIIK_Enter);
}
/* WaitForReturn */

/******************************************************************************/

static void UpdateDVIStatusLine (void)
{
  /* Show totalpages, currDVIpage, currTeXpage, direction and displaymode. */

  int i, lastnonzero;

  ClearTextLine  (DVIstatusl);
  MoveToTextLine (DVIstatusl);

  MesgInt (totalpages);
  MesgString (" pages");

  MesgString (" DVI page=");
  MesgInt (currDVIpage);

  MesgString (" TeX page=");

  MesgChar ('[');

  lastnonzero = 9;
  while (lastnonzero > 0  &&  currTeXpage[lastnonzero] == 0)
  {
    lastnonzero--;   /* find last counter with non-zero value */
  }
  /* always show \count0 but don't show trailing 0 counters */
  for (i = 0; i <= lastnonzero; i++)
  {
    MesgInt (currTeXpage[i]);

    if (i != lastnonzero)
    {
      MesgChar ('.');
    }
  }
  MesgChar (']');

  if (displaymode & boxmode)
    MesgString ("B");
  else
    MesgString (" ");

  MesgLine();
}
/* UpdateDVIStatusLine */


/******************************************************************************/

static void WriteDimension (double precision, double res, int pixels)
{
  /* Show the given pixel dimension in terms of currentunits. */

  double realdim=0;
  int fracpart;

  switch (currentunits)
  {
  case ic:
    realdim = (double) pixels / res;
    break;

  case cm:
    realdim = (double) pixels / res * 2.54;
    break;

  case mm:
    realdim = (double) pixels / res * 25.4;
    break;

  case bp:
    realdim = (double) pixels / res * 72.0;
    break;

  case pc:
    realdim = (double) pixels / res * 72.27 / 12.0;
    break;

  case pt:
    realdim = (double) pixels / res * 72.27;
    break;

  case sp:
    realdim = (double) pixels / res * 72.27 * 65536.0;
    break;

  case px:
    MesgInt (pixels);
    break;
  } /* currentunits */

  if (currentunits != px)
  {
    /* gt - This seems immensely complicated, for what it's doing! */
    /* show realdim to the specified precision */
    if (fabs (realdim) < 0.5 * precision)
    {
      MesgString ("0.0");
    }
    else
    {
      if (realdim < 0.0)
      {
        MesgChar ('-');
        realdim = fabs (realdim);
      }
      realdim += 0.5 * precision;   /* round up to specified precision */
      MesgInt ((int) realdim);   /* whole part */
      MesgChar ('.');
      fracpart = (int) ((realdim - (int) realdim) / precision);
      MesgInt (fracpart);
    } /* if */
  } /* if */
}
/* WriteDimension */

/******************************************************************************/

static void WriteXDim (double precision, int pixels)
{
  WriteDimension (precision, xres, pixels);
}
/* Write XDim */

/******************************************************************************/

static void WriteYDim (double precision, int pixels)
{
  WriteDimension (precision, yres, pixels);
}
/* Write YDim */

/******************************************************************************/

static void UpdateWindowStatusLine (void)
{
  /* Show current window location and size, page location and size, and units. */

  double precision = 0.1;  /* precision for status line */

  ClearTextLine (windowstatusl);
  MoveToTextLine (windowstatusl);

  MesgString ("Window at (");
  WriteXDim (precision, windowleft);
  MesgChar (',');
  WriteYDim (precision, windowtop);
  MesgString (") ");
  WriteXDim (precision, scaledwd);
  MesgString (" by ");
  WriteYDim (precision, scaledht);

  MesgString ("  Page at (");
  WriteXDim (precision, minhp);
  MesgChar (',');
  WriteYDim (precision, minvp);
  MesgString (") ");
  WriteXDim (precision, maxhp - minhp + 1);
  MesgString (" by ");
  WriteYDim (precision, maxvp - minvp + 1);
  MesgString ("  ");

  switch (currentunits)
  {
  case ic:
    MesgString ("IN");
    break;

  case cm:
    MesgString ("CM");
    break;

  case mm:
    MesgString ("MM");
    break;

  case bp:
    MesgString ("BP");
    break;

  case pc:
    MesgString ("PC");
    break;

  case pt:
    MesgString ("PT");
    break;

  case sp:
    MesgString ("SP");
    break;

  case px:
    MesgString ("PX");
    break;
  } /* currentunits */
  MesgLine();
}
/* UpdateWindowStatusLine */


/******************************************************************************/

#define limit           2147483647
/* TeX's limit = 2^31 - 1.
                                      Should also be >= maxpix.
                                      Note that this also defines the range of
                                      page numbers the user can ask for! */
#define threshold       (limit / 10)   /* nearing overflow */


static boolean GetInteger (Char * str_in, int slen, int * pos, int * n)
{
  /* Extract an integer from given str_in starting at given pos.
     pos is also used to return the position after the integer.
     If no integer is found, then set n to 0 and return FALSE
     (in that case, pos will only change if leading spaces were skipped).
     If ABS(n) > limit then set n to sign * limit.
     Valid syntax is  +{digit}  or  -{digit}  or  digit{digit}.
     Note that a + or - by itself is valid and sets n to 0.
  */

  string str;

  int absval, last, sign;
  boolean inttoobig;
  boolean intfound = false;

  (void) memcpy (str, str_in, sizeof (string));

  while (*pos < slen  && str[*pos] == ' ')  /* skip any leading spaces */
  {
    (*pos)++;
  }

  absval = 0;
  sign = 1;
  last = *pos;
  inttoobig = false;

  if (*pos < slen)
  {
    if (str[*pos] == '-')
    {
      sign = -1;
      last++;
    }
    else
    {
      if (str[*pos] == '+')
        last++;
    }

    while (last < slen  &&  (str[last] >= '0' && str[last] <= '9'))
    {
      if (absval > threshold || (absval == threshold && str[last] > '7'))
        inttoobig = true;
      else
        absval = absval * 10 + str[last] - '0';
      last++;
    }
  }

  if (*pos == last)
  {
    *n = 0;
    intfound = false;
  }
  else
  {
    *pos = last;
    if (inttoobig)
      absval = limit;
    *n = sign * absval;
    intfound = true;
  }

  return intfound;
}  /* GetInteger */

#undef limit
#undef threshold

/******************************************************************************/

static boolean GetDimension (double res, Char * str_in, int slen,
                             int * pos, int * n)
{
  /* Extract a dimension from given str_ starting at given pos.
     n returns the corresponding number of pixels in the dimension
     (which is an integer or real value in terms of currentunits);
     pos is also used to return the position after the dimension.
     If no dimension is found then set n to 0 and return FALSE (pos will only
     change if leading spaces were skipped).
     If ABS(n) > maxpix then set n to sign * maxpix.
     Valid syntax of a dimension is  integer[.{digit}]  or  .{digit}  where
     an integer is defined by GetInteger.
     Real dimensions are truncated to 4 decimal places.
     Note that a sign or decimal point by itself is valid and sets n to 0.
  */

  boolean Result;
  string str;

  int sign, intdim, fracpart, divisor;
  double absrealdim;
  boolean intpresent, dimtoobig;

  (void) memcpy (str, str_in, sizeof (string));

  /* GetInteger does not remember a sign by itself, so we need to check
     for -ve dimensions like -.5 first.
  */

  while (*pos < slen  &&  str[*pos] == ' ')  /* skip any leading spaces */
  {
    (*pos)++;
  }

  sign = 1;
  if (*pos < slen)
  {
    if (str[*pos] == '-')
      sign = -1;
  }
  intpresent = GetInteger (str, slen, pos, &intdim);
  if (!intpresent)
  {
    if (*pos == slen || str[*pos] != '.')
    {
      *n = 0;
      Result = false;
      goto _L999;

    }
  } /* if */

  /* dimension is valid; if no integer part then intdim will be 0; sign = +|-1 */
  if (*pos == slen || str[*pos] != '.')
  {
    /* no fractional part */
    absrealdim = abs (intdim);
  }
  else
  {
    /* extract fractional part */
    (*pos)++;   /* skip over decimal point */
    divisor = 1;
    fracpart = 0;
    while (*pos < slen  &&  (str[*pos] >= '0' && str[*pos] <= '9'))
    {
      /* only consider up to 4 decimal places */
      if (divisor < 10000)
      {
        divisor *= 10;
        fracpart = fracpart * 10 + str[*pos] - '0';
      }
      (*pos)++;
    } /* while */
    absrealdim = abs (intdim) + (double)fracpart / divisor;
  } /* if */

  /* calculate n based on absrealdim, sign and currentunits */
  dimtoobig = false;
  switch (currentunits)
  {
    case ic:
      if (absrealdim > (double)maxpix / res)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim * res + 0.5);
      break;

    case cm:
      if (absrealdim > (double)maxpix / res * 2.54)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim / 2.54 * res + 0.5);
      break;

    case mm:
      if (absrealdim > (double)maxpix / res * 25.4)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim / 25.4 * res + 0.5);
      break;

    case bp:
      if (absrealdim > (double)maxpix / res * 72.0)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim / 72.0 * res + 0.5);
      break;

    case pc:
      if (absrealdim > (double)maxpix / res * (72.27 / 12.0))
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim / 72.27 * 12.0 * res + 0.5);
      break;

    case pt:
      if (absrealdim > (double)maxpix / res * 72.27)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim / 72.27 * res + 0.5);
      break;

    case sp:
      if (absrealdim > (double)maxpix / res * 72.27 * 65536.0)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim / (72.27 * 65536.0) * res + 0.5);
      break;

    case px:
      if (absrealdim > maxpix)
	dimtoobig = true;
      else
	*n = sign * (int)(absrealdim + 0.5);
      break;
  } /* currentunits */

  if (dimtoobig)
    *n = sign * maxpix;

  Result = true;

_L999: ;
  return Result;
}
/* GetDimension */

/******************************************************************************/

static boolean GetXDim (Char * str_in, int slen, int * pos, int * n)
{
  return GetDimension (xres, str_in, slen, pos, n);
}
/* GetXDim */

/******************************************************************************/

static boolean GetYDim (Char * str_in, int slen, int * pos, int * n)
{
  return GetDimension (yres, str_in, slen, pos, n);
}
/* GetYDim */

/******************************************************************************/
/****************************************************************************/

Inline static int Min (int a, int b)
{
  /* Return the minimum value of a and b. */

  if (a < b)
    return a;
  else
    return b;
}
/* Min */

/****************************************************************************/

Inline static int Max (int a, int b)
{
  /* Return the maximum value of a and b. */

  if (a > b)
    return a;
  else
    return b;
}
/* Max */

/*****************************************************************************/

static void Valid ( )
{
  /* Reset the invalid region. */ 

  invalidtop = MAXINT;
  invalidbottom = 0;
  invalidleft = MAXINT;
  invalidright = 0;
} 
/* Valid */

/*****************************************************************************/

static void Invalid (int top, int left, int bottom, int right)
{
  /* Add the region defined by top, left, bottom, and right to the invalid
     region. */

/*  invalidtop = Min(invalidtop, top);
  invalidbottom = Max(invalidbottom, bottom);
  invalidleft = Min(invalidleft, left);
  invalidright = Max(invalidright, right);*/

  invalidtop = top;
  invalidbottom = bottom;
  invalidleft = left;
  invalidright = right;
} 
/* Invalid */

/*****************************************************************************/

Inline static int ScaleHpos (int h)
{
  /* Return a scaled value for the given horizontal window coordinate. */

  if (hscalefactor > 1.0)
    return ((int)(h * hscalefactor + 0.5));
  else   /* hscalefactor <= 1.0 */
    return ((int)((h + 0.5) * hscalefactor));
}
/* ScaleHpos */

/*****************************************************************************/

__inline__ static int ScaleVpos (int v)
{
  /* Return a scaled value for the given vertical window coordinate. */

  if (vscalefactor > 1.0)
    return ((int)(v * vscalefactor + 0.5));
  else   /* vscalefactor <= 1.0 */
    return ((int)((v + 0.5) * vscalefactor));
}
/* ScaleVpos */

/*****************************************************************************/

static void NewLocation (int newhp, int newvp)
{
  /* Change window location to given position and update window edges.
     If entire window moves outside non-empty page rectangle then outsidepage
     becomes TRUE and we restrict movement to just beyond the edge(s) so that
     user can easily move window (via Up,Down,Left,Right) to positions
     in which one or more window and page edges coincide.
     Note that allpagevisible is also updated.
  */

  int x1,y1,x2,y2;

  outsidepage = false;
  if (currDVIpage != 0 && !pageempty)
  {
    /* check if new position puts window entirely outside edges;
       if so then minimize the movement needed to keep this true */

    if (newvp > maxvp)
    {
      outsidepage = true;
      newvp = maxvp + 1;
    }
    else if (newvp <= minvp - scaledht)
    {
      outsidepage = true;
      newvp = minvp - scaledht;
    }

    if (newhp > maxhp)
    {
      outsidepage = true;
      newhp = maxhp + 1;
    }
    else if (newhp <= minhp - scaledwd)
    {
      outsidepage = true;
      newhp = minhp - scaledwd;
    }
  }

  y1 = newvp;
  x1 = newhp;
  y2 = y1 + scaledht - 1;
  x2 = x1 + scaledwd - 1;
  Invalid(y1,x1,y2,x2);

  /* Set new screen position */

  windowtop = newvp;
  windowleft = newhp;
  windowbottom = windowtop + scaledht - 1;
  windowright = windowleft + scaledwd - 1;
  allpagevisible = (currDVIpage != 0 && !pageempty
                    && (minvp >= windowtop  && maxvp <= windowbottom)
                    && (minhp >= windowleft && maxhp <= windowright));

  /* even if pageempty or window hasn't moved we must still call DisplayPage */

  if (currDVIpage != 0)
    paintwindow = true;
}
/* NewLocation */

static void WindowUpDown (void)
{
  int amount = scaledht/2;  /* for odd scaleht, this is half a pixel less */
  if (command==UpCmd) amount = -amount;
  NewLocation (windowleft, windowtop + amount);
}
/* WindowUpDown */

/*****************************************************************************/

static void WindowLeftRight (void)
{
  int amount = scaledwd/2;  /* for odd scaledwd, this is half a pixel less */
  if(command==LeftCmd) amount = -amount;
  NewLocation (windowleft + amount, windowtop);
}
/* WindowLeftRight */

/*****************************************************************************/

static void NewWindowWidth (int wd)
{
  /* Set window width to given value (> 0 and <= max dimension). */

  scaledwd = wd;
  hscalefactor = (double) windowwd / scaledwd;
}
/* NewWindowWidth */

/******************************************************************************/
/******************************************************************************/

static void NewWindowHeight (int ht)
{
  /* Set window height to given value (> 0 and <= max dimension). */

  scaledht = ht;
  vscalefactor = (double)windowht / scaledht;
}
/* NewWindowHeight */

/******************************************************************************/
/*****************************************************************************/
/******************************************************************************/

static void ZITL (void)
{
  NewWindowWidth  ((int) ceil (scaledwd / 2.0));
  NewWindowHeight ((int) ceil (scaledht / 2.0));
}

/******************************************************************************/

static void ZOTL (void)
{
  if (maxpix / 2.0 > scaledwd) NewWindowWidth ((int) (scaledwd * 2.0));
  else NewWindowWidth (maxpix);
  if (maxpix / 2.0 > scaledht) NewWindowHeight ((int) (scaledht * 2.0));
  else NewWindowHeight (maxpix);
}

/******************************************************************************/

static void ZoomWindow (void)
{
    paintDVIStatus = true;

    if (command==ZoomInCmd)
    {
	NewLocation (windowleft + scaledwd/2, windowtop + scaledht/2);
	ZITL();
	NewLocation (windowleft - scaledwd/2, windowtop - scaledht/2);
    }
    else
    {
	NewLocation (windowleft + scaledwd/2, windowtop + scaledht/2);
	ZOTL();
	NewLocation (windowleft - scaledwd/2, windowtop - scaledht/2);
    }
}

/******************************************************************************/

static boolean NextPageFound (boolean ascending)
{
  /* User has selected next page in DVI file; what they get will depend on
     the current DVI page and whether we are ascending or not.
     Return TRUE iff we can move to next page.
  */

  boolean found = false;

  if (currDVIpage == 1 && !ascending)
  {
    ClearMessageLine();
    MesgString ("You are looking at first DVI page!");
    found = false;
  }
  else if (currDVIpage == totalpages && ascending)
  {
    ClearMessageLine();
    MesgString ("You are looking at last DVI page!");
    found = false;
  }
  else
  {
    MoveToNextPage (ascending);   /* position to next DVI page */
    found = true;
  }
  return found;
}
/* NextPageFound */

/******************************************************************************/

static void DVIPageFound (int n)
{
  /* User has selected a particular DVI page number.
     Move to page n and return TRUE iff n is in 1..totalpages.
  */

  if(n<1) n=totalpages;
  if(n>totalpages) n=totalpages;
  MoveToDVIPage (n);   /* position to given DVI page */
}
/* DVIPageFound */

/*****************************************************************************/

static void ProcessPage (void)
{
  /* We are ready to interpret the current DVI page and fill in the various
     data structures imported from DVIReader.  This routine will also: set the
     window size and location to useful values (if autoviewing), update
     pageoffpaper (after checking to see if it was TRUE for the previous page
     processed as part of a multiple command string), set screenjustcleared,
     paintwindow and paintWindowStatus to TRUE, set paintDVIStatus to
     FALSE. */

  int halfht, halfwd;

  /* We check pageoffpaper here so user can type "NNNNNNNNNNNN..." and note
     ALL the pages that are off the paper, not just the last one processed. */
  if (pageoffpaper)
  {
    ClearMessageLine();
    MesgString ("Page off paper!");
        /* the previous page */
    WaitForReturn();
  }
  ClearScreen();
  screenjustcleared = true;
  UpdateDVIStatusLine();
      /* a MoveTo... routine has updated currDVI/TeXpage */
  paintDVIStatus = false;
  InterpretPage();   /* fill in DVIReader's page data structures */
  SortFonts(&unusedfont);
  /* sort fonts in order of least chars and return
                                pointer to first unused font */
  ClearMessageLine();   /* clear any message */
  if (pageempty)
  {
    minhp = 0;
    maxhp = 0;
    minvp = 0;
    maxvp = 0;   /* for window status */
  }

  /* view as much of paper as possible, but without too much distortion */
    if ((paperwd <  paperht && windowwd >= windowht)
     || (paperwd == paperht && windowwd >  windowht))
    {
      halfht = paperht / 2;
      if (paperht & 1)   /* ensure bottom visible */
        halfht++;
      NewWindowHeight (halfht);   /* try top half of paper */
      NewWindowWidth (paperwd);
      NewLocation (paperleft, papertop);   /* top left corner of paper */
      if (!pageempty && outsidepage)   /* try moving down */
        NewLocation (paperleft, papertop + halfht);
    }
    else if ((paperwd >  paperht && windowwd <= windowht)
          || (paperwd == paperht && windowwd <  windowht))
    {
      halfwd = paperwd / 2;
      if (paperwd & 1)   /* ensure right visible */
        halfwd++;
      NewWindowHeight (paperht);
      NewWindowWidth (halfwd);   /* try left half of paper */
      NewLocation (paperleft, papertop);   /* top left corner of paper */
      if (!pageempty && outsidepage)   /* try moving right */
        NewLocation (paperleft + halfwd, papertop);
    }
    else
    {
      /* paper shape matches unscaled window shape */
      NewWindowHeight (paperht);   /* try all of paper */
      NewWindowWidth (paperwd);
      NewLocation (paperleft, papertop);   /* top left corner of paper */
    }

  /* check if part/all of page is off paper;
     if so, and autoviewing is enabled, then we set window size and location
     so user can just see ALL of paper AND ALL of page.
  */
  pageoffpaper = (!pageempty && ((minhp < paperleft  || minvp < papertop)
                              || (maxhp > paperright || maxvp > paperbottom)));
  if (pageoffpaper)
  {
    NewWindowHeight (Max (maxvp, paperbottom) - Min (minvp, papertop)  + 1);
    NewWindowWidth  (Max (maxhp, paperright)  - Min (minhp, paperleft) + 1);
    NewLocation (Min (minhp, paperleft), Min (minvp, papertop));
  }
  Invalid(windowtop, windowleft, windowbottom, windowright);
  paintWindowStatus = true;
}
/* ProcessPage */

/******************************************************************************/

static void WriteUnits (void)
{
  switch (currentunits)
  {
  case ic:
    MesgString ("in");
    break;

  case cm:
    MesgString ("cm");
    break;

  case mm:
    MesgString ("mm");
    break;

  case bp:
    MesgString ("bp");
    break;

  case pc:
    MesgString ("pc");
    break;

  case pt:
    MesgString ("pt");
    break;

  case sp:
    MesgString ("sp");
    break;

  case px:
    MesgString ("px");
    break;
  } /* currentunits */
}
/* WriteUnits */

/******************************************************************************/

static int WritePtSize (int scaledsize)
{
  /* Show given font size (in DVI units), in (possibly magnified) pts.
     Return length of displayed message.
  */

  double realdim;
  int fracpart;
  double precision = 0.1;
  int len = 0;

  realdim = scaledsize / 65536.0 * (mag / 1000.0);
  /* show realdim to the precision given above */
  if (fabs (realdim) < 0.5 * precision)
  {
    MesgChar ('0');
    ++ len;
  }
  else
  {
    if (realdim < 0.0)
    {
      MesgChar ('-');
      ++ len;
      realdim = fabs (realdim);
    }
    realdim += 0.5 * precision;   /* round up to specified precision */
    len += MesgInt ((int) realdim);   /* whole part */
    fracpart = (int) ((realdim - (int) realdim) / precision);
    if (fracpart > 0)
    {
      MesgChar ('.');
      ++ len;
      len += MesgInt (fracpart);
    }
  }
  MesgString ("pt");
  len += 2;
  return (len);
}
/* WritePtSize */

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

#define maxline         80
    /* SYSDEP: cmd_help should have <= maxline chars/line */

/******************************************************************************/

static boolean PrintText (Char * * text)
{
  int retval = false;

  if (text == (Char * *) NULL)
  {
    retval = false;   /* No text to print. */
  }
  else
  {
    /* Print each text string */
    for (;  *text != (char *) NULL;  text++)
    {
      printf ("%s", *text);
      /* If there's no newline in *text, print one. */
      if (strchr (*text, '\n') == (char *) NULL)
        printf ("\n");
    }
    retval = true;
  }
  return retval;
}
/* PrintText */

/******************************************************************************/

#ifdef __STDC__
static boolean PrintFile (Char * filename)
#else
static boolean PrintFile (filename)
    Char * filename;
#endif
{
  int retval = false;
  FILE * filep = (FILE *) NULL;    /* file pointer for "filename" */

  filep = fopen (filename, "r");   /* read only */
  if (filep == (FILE *) NULL)
  {
    /* File was not opened. */
    retval = false;
  }
  else
  {
    Char outline[maxline + 1];        /* Allow space for terminating null */

    while (fgets (outline, maxline+1, filep) != (char *) NULL)
    {
      printf ("%s", outline);
      /* If there's no newline in cp, print one. */
      if (strchr (outline, '\n') == (char *) NULL)
        printf ("\n");
    }
    retval = true;
  }
  return retval;
}
/* PrintFile */

/******************************************************************************/
/*****************************************************************************/

#undef maxline

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/******************************************************************************/

#define edgepixel       '.'
    /* black pixel for outlines on non-graphic VDUs;
       note that VDU TeXtoASCII['.'] = '.' */

static void DisplayPaperEdges (void)
{
  /* Display visible outlines of the imaginary sheet of paper.
     Thickness of outlines = 1 screen pixel,
     no matter what the h and v scaling.
  */

  int top, bot, left, right;   /* visible edges of paper in paper pixels */
  int scaledtop, scaledleft;   /* scaled visible edges in screen pixels */
  int scaledbot, scaledright, scaledheight;
  /* scaled width and height */
  int scaledwidth;

  /* first check if any part of paper is visible */

  if (papertop > windowbottom)
    goto _L999;

  if (paperbottom < windowtop)
    goto _L999;

  if (paperleft > windowright)
    goto _L999;

  if (paperright < windowleft)
    goto _L999;

  /* part or all of paper is visible, so return visible region */
  top = Max (papertop, windowtop);
  bot = Min (paperbottom, windowbottom);
  left = Max (paperleft, windowleft);
  right = Min (paperright, windowright);
  scaledtop = ScaleVpos (top - windowtop) + windowv;
  scaledleft = ScaleHpos (left - windowleft) + windowh;
  if (vscalefactor > 1.0)
    scaledbot = ScaleVpos (bot - windowtop + 1) + windowv - 1;
  else
    scaledbot = ScaleVpos (bot - windowtop) + windowv;
  if (hscalefactor > 1.0)
    scaledright = ScaleHpos (right - windowleft + 1) + windowh - 1;
  else
    scaledright = ScaleHpos (right - windowleft) + windowh;
  scaledheight = scaledbot - scaledtop + 1;
  scaledwidth = scaledright - scaledleft + 1;
  /* only show visible edges if they are also paper outlines */
  if (left == paperleft)
    ShowRectangle (scaledleft, scaledtop, 1, scaledheight, edgepixel);
  if (bot == paperbottom)
    ShowRectangle (scaledleft, scaledbot, scaledwidth, 1, edgepixel);
  if (top == papertop)
    ShowRectangle (scaledleft, scaledtop, scaledwidth, 1, edgepixel);
  if (right == paperright)
    ShowRectangle (scaledright, scaledtop, 1, scaledheight, edgepixel);

_L999: ;
}
/* DisplayPaperEdges */

#undef edgepixel


/******************************************************************************/

#ifdef __STDC__
static boolean RectangleVisible (int intop, int inbot, int inleft, int inright,
  int *outtop, int *outbot, int *outleft, int *outright)
#else
static boolean RectangleVisible (intop, inbot, inleft, inright,
  outtop, outbot, outleft, outright)
int intop; int inbot; int inleft; int inright;
  int *outtop; int *outbot; int *outleft; int *outright;
#endif
{
  /* Return TRUE iff part or all of given rectangle would be visible
     in the current window.  Iff so, then we also return the visible
     region; the input and possible output rectangles are defined by their
     top, bottom, left and right edges in paper pixel coordinates.
  */

  if (intop > invalidbottom)
    return false;
  else if (inbot < invalidtop)
    return false;
  else if (inleft > invalidright)
    return false;
  else if (inright < invalidleft)
    return false;
  else
  {
    /* part or all of rectangle is visible, so return visible region */
    *outtop = Max (intop, invalidtop);
    *outbot = Min (inbot, invalidbottom);
    *outleft = Max (inleft, invalidleft);
    *outright = Min (inright, invalidright);
    return true;
  }
}
/* RectangleVisible */

/*****************************************************************************/

#define rulepixel       '*'
    /* black pixel for rules on non-graphic VDUs;
       note that VDU sets TeXtoASCII['*'] := '*' */


static void DisplayRules (void)
{
  /* Display all pixels in rules, regardless of current displaymode.
     Rules will be displayed in the same order as in the DVI page (essentially
     top-down and left-right) because of the way DVIReader builds a rulelist.
  */

  int top, bottom, left, right;   /* visible edges of rule */
  int scaledtop, scaledleft;   /* scaled visible edges */
  int scaledbot, scaledright, scaledwidth, scaledheight;
      /* scaled width and height */
  int thisrule;
  _REC_ruletable * ruletab = (_REC_ruletable *) NULL;

  thisruleinfo = rulelist;
  while (thisruleinfo != (ruleinfo *) NULL)
  {
    thisrule = 0;
    while (thisrule < thisruleinfo->rulecount)
    {
      ruletab = &thisruleinfo->ruletable[thisrule];
      /* check if any part of rule is visible */
      /* vp,hp is bottom left corner of rule on page */
      if (RectangleVisible (ruletab->vp - ruletab->ht + 1,
                            ruletab->vp, ruletab->hp,
                            ruletab->hp + ruletab->wd - 1,
                            &top, &bottom, &left, &right))
      {   /* rule edges */
        /* show all pixels in this rectangle */
        scaledtop = ScaleVpos (top - windowtop) + windowv;
        scaledleft = ScaleHpos (left - windowleft) + windowh;

        if (vscalefactor > 1.0)
          scaledbot = ScaleVpos (bottom - windowtop + 1) + windowv - 1;
        else
          scaledbot = ScaleVpos (bottom - windowtop) + windowv;

        if (hscalefactor > 1.0)
          scaledright = ScaleHpos (right - windowleft + 1) + windowh - 1;
        else
          scaledright = ScaleHpos (right - windowleft) + windowh;

        /* v coord of top left cnr */
        scaledheight = scaledbot - scaledtop + 1;

        /* h coord of top left cnr */
        scaledwidth = scaledright - scaledleft + 1;

        ShowRectangle (scaledleft, scaledtop, scaledwidth, scaledheight,
                       rulepixel);
      }  /* if */
      /* visible rectangle */
      thisrule++;
    }  /* while */
    thisruleinfo = thisruleinfo->nextrule;
  }  /* while */
}
/* DisplayRules */

#undef rulepixel


/******************************************************************************/

#ifdef __STDC__
static boolean PixelVisible (int hpos, int vpos)
#else
static boolean PixelVisible (hpos, vpos)
int hpos; int vpos;
#endif
{
  /* Return TRUE iff given paper pixel would be visible in current window. */

  if (allpagevisible)
    return true;
  else if (vpos < windowtop)
    return false;
  else if (vpos > windowbottom)
    return false;
  else if (hpos < windowleft)
    return false;
  else if (hpos > windowright)
    return false;
  else
    return true;
}
/* PixelVisible */

/******************************************************************************/

static boolean TerseChar (void)
{
  /* Display a quick and nasty representation of character,
     only if ref pt visible.
     Just how good the representation is,
     depends on the capabilities of the VDU.
     We don't bother checking
     whether glyph is actually blank or non-existent.
  */

  _REC_chartable * chartab = &thischarinfo->chartable[thischar];

  if (chartab == (_REC_chartable *) NULL)
    return false;

  if (PixelVisible (chartab->hp, chartab->vp))   /* ref pt of char is visible */
    ShowChar (ScaleHpos (chartab->hp - windowleft) + windowh,
              ScaleVpos (chartab->vp - windowtop)  + windowv,
              chartab->code);

  return true;
}
/* TerseChar */

/******************************************************************************/

static boolean BoxChar (void)
{
  /* Display visible box outlines of glyph.
     Thickness of outlines = 1 screen pixel,
     no matter what the h and v scaling.
  */

  int vpmyo, hpmxo;   /* vp-yo, hp-xo: glyph's top and left edges */
  int top, bottom, left, right;   /* visible edges of glyph */
  int scaledtop, scaledleft;   /* scaled visible edges */
  int scaledbot, scaledright, scaledheight;
  /* scaled width and height */
  int scaledwidth;
  Char ch;

  _REC_chartable * chartab = &thischarinfo->chartable[thischar];
  _REC_pixeltable * pixtab = (_REC_pixeltable *) NULL;

  if (chartab == (_REC_chartable *) NULL)
    return false;

  pixtab = &thisfontinfo->pixelptr[chartab->code];
  if (pixtab == (_REC_pixeltable *) NULL)
    return false;

  if (pixtab->mapadr != 0)   /* glyph present and non-blank */
  {
    /* check if any part of glyph is visible */

    vpmyo = chartab->vp - pixtab->yo;
    hpmxo = chartab->hp - pixtab->xo;

    if (RectangleVisible (vpmyo, vpmyo + pixtab->ht - 1,
			  hpmxo, hpmxo + pixtab->wd - 1,
			  &top, &bottom, &left, &right))
    {   /* glyph edges */

      scaledtop  = ScaleVpos (top - windowtop)   + windowv;
      scaledleft = ScaleHpos (left - windowleft) + windowh;

      if (vscalefactor > 1.0)
	scaledbot = ScaleVpos (bottom - windowtop + 1) + windowv - 1;
      else
	scaledbot = ScaleVpos (bottom - windowtop) + windowv;

      if (hscalefactor > 1.0)
	scaledright = ScaleHpos (right - windowleft + 1) + windowh - 1;
      else
	scaledright = ScaleHpos (right - windowleft) + windowh;

      scaledheight = scaledbot   - scaledtop + 1;
      scaledwidth  = scaledright - scaledleft + 1;

      /* Only show edges that are also glyph outlines!
	 Following method reduces the number of ShowRectangle calls needed for
	 very small boxes.
      */
      ch = chartab->code;
      if ((scaledheight < 3
           && (top  == vpmyo && bottom == vpmyo + pixtab->ht - 1))
       || (scaledwidth  < 3
           && (left == hpmxo && right  == hpmxo + pixtab->wd - 1)))
      {
	ShowRectangle (scaledleft, scaledtop, scaledwidth, scaledheight, ch);
      }
      else
      {
	if (left == hpmxo)
	  ShowRectangle (scaledleft, scaledtop, 1, scaledheight, ch);
	if (bottom == vpmyo + pixtab->ht - 1)
	  ShowRectangle (scaledleft, scaledbot, scaledwidth, 1, ch);
	if (top == vpmyo)
	  ShowRectangle (scaledleft, scaledtop, scaledwidth, 1, ch);
	if (right == hpmxo + pixtab->wd - 1)
	  ShowRectangle (scaledright, scaledtop, 1, scaledheight, ch);
      }
    }
    /* visible part */
  }

  return true;
}
/* BoxChar */

/******************************************************************************/

#ifdef FULL_CHAR_PS
static boolean FullCharPS (void)
{
  /* Display filled rectangle approximating the extent of PostScript glyph. */

  int vpmyo, hpmxo;   /* vp-yo, hp-xo: glyph's top and left edges */
  int top, bottom, left, right;   /* visible edges of glyph */
  int scaledtop, scaledleft;   /* scaled visible edges */
  int scaledbot, scaledright, scaledheight;
  /* scaled width and height */
  int scaledwidth;

  _REC_chartable * chartab = (_REC_chartable *) NULL;
  _REC_pixeltable * pixtab = (_REC_pixeltable *) NULL;

  chartab = &thischarinfo->chartable[thischar];
  if (chartab == (_REC_chartable *) NULL)
    return false;

  pixtab = &thisfontinfo->pixelptr[chartab->code];
  if (pixtab == (_REC_pixeltable *) NULL)
    return false;

  if (pixtab->mapadr != 0)   /* glyph present and non-blank */
  {
    /* check if any part of glyph is visible */

    vpmyo = chartab->vp - pixtab->yo;
    hpmxo = chartab->hp - pixtab->xo;

    if (RectangleVisible (vpmyo, vpmyo + pixtab->ht - 1,
                          hpmxo, hpmxo + pixtab->wd - 1,
                          &top, &bottom, &left, &right))
    {   /* glyph edges */

      scaledtop  = ScaleVpos (top - windowtop)   + windowv;
      scaledleft = ScaleHpos (left - windowleft) + windowh;

      if (vscalefactor > 1.0)
	scaledbot = ScaleVpos (bottom - windowtop + 1) + windowv - 1;
      else
	scaledbot = ScaleVpos (bottom - windowtop) + windowv;

      if (hscalefactor > 1.0)
	scaledright = ScaleHpos (right - windowleft + 1) + windowh - 1;
      else
	scaledright = ScaleHpos (right - windowleft) + windowh;

      scaledheight = scaledbot - scaledtop + 1;

      scaledwidth = scaledright - scaledleft + 1;

      ShowRectangle (scaledleft, scaledtop, scaledwidth, scaledheight,
		    chartab->code);
    }
    /* visible part */
  }

  return true;
}
/* FullCharPS */
#endif /* FULL_CHAR_PS */

/******************************************************************************/

#ifdef __STDC__
static void NotFound (Char *fspec)
#else
static void NotFound (fspec)
Char *fspec;
#endif
{
  StartText();
  ResetVDU();   /* do before message since it might erase screen! */

  MesgString ("Couldn't open font ");
  MesgString (fspec);
  MesgChar ('!');
  MesgLine();

  RestoreTerminal();
  exit (1);
}
/* NotFound */

/******************************************************************************/

#define MAXVISWORDS     100
#define WORDSIZE  32

/* MAXVISWORDS * WORDSIZE = 100 * 32 = 3200 bits = maximum pixel width
   of glyph!
   If any fonts have glyphs wider than this, then increase MAXVISWORDS.
*/

/* SYSDEP: BITSET is 32 bit word with elements 31,30,29,...,0 */
typedef Word glyphrow[MAXVISWORDS];


static boolean FullChar (void)
{
  /* Display all pixels in a glyph using bitmap from font file.
     The algorithm avoids overlapping rows when vscalefactor < 1.0.
     When hscalefactor < 1.0, it is not worth the extra code to avoid
     overlapping runs of 1 bits because the majority of character glyphs
     have only one or two runs per row.
  */

  int vpmyo, hpmxo;   /* vp-yo, hp-xo: glyph's top and left edges */
  int top, bottom, left, right;   /* visible edges of glyph */
  int scaledv, scalednextv;   /* scaled vertical positions for rows */
  int scaledh=0;   /* scaled horizontal positions within row */
  int scaledwidth, scaledheight;   /* scaled width and height of row */
  int thisrow, thisbit;   /* in paper coordinates */
  int wordsperrow;   /* rows of bitmap are word aligned */
  int firstbit, lastbit;   /* somewhere in 0 .. wordsperrow*WORDSIZE-1 */
  int firstword, lastword;   /* somewhere in 0 .. wordsperrow-1 */
  int endword;   /* = visible words in row, - 1 */
  int wordpos;   /* 0 .. endword */
  int bitpos;    /* (WORDSIZE-1) .. 0 */
  int i;

  glyphrow row;  /* holds VISIBLE bits in one row of glyph;
                    possibly > one row if vscalefactor < 1.0 */

  int_or_bptr ptr;   /* pointer into bitmap - actually a "struct" */
  int inrun;   /* are we in a run of black pixels in row? */

  _REC_chartable * chartab = &thischarinfo->chartable[thischar];
  _REC_pixeltable * pixtab = &thisfontinfo->pixelptr[chartab->code];

  if (pixtab->mapadr != 0) {  /* glyph present and non-blank */
    /* check if any part of glyph is visible */

    vpmyo = chartab->vp - pixtab->yo;
    hpmxo = chartab->hp - pixtab->xo;
    if (RectangleVisible (vpmyo, vpmyo + pixtab->ht - 1, hpmxo,
			  hpmxo + pixtab->wd - 1, 
			  &top, &bottom, &left, &right)) {   /* glyph edges */
      if (pixtab->bitmap.UU.mptr == (Word *) NULL) {
	if (!fontopen) {   /* dimensions of bitmap */
	  if (thisfontinfo->fontexists) {
	    if (!OpenFontFile (thisfontinfo->fontspec))
	      NotFound (thisfontinfo->fontspec);
	  } else {
	    string dummyspec;
	    BuildFileSpec (pkdir, dummy_pk, dummyspec);
	    if (!OpenFontFile (dummyspec))
	      NotFound (dummyspec);
	  }
	  fontopen = true;   /* only open font once */
	}
	/* bitmap info in font file */
	    /* &pixtab->bitmap receives the starting address of bitmap */
	if (!GetBitmap (pixtab->ht, pixtab->wd, pixtab->mapadr, &pixtab->bitmap))
	  return false;
      }
      /* Words in 1 row of bitmap */
      wordsperrow = (pixtab->wd + (WORDSIZE-1)) / WORDSIZE;

      firstbit = left - hpmxo;   /* first visible bit */
      lastbit = right - hpmxo;   /* last visible bit */
      firstword = firstbit / WORDSIZE;   /* first visible word */
      lastword = lastbit / WORDSIZE;   /* last visible word */
      endword = lastword - firstword;

      /* set the visible words in row to 0 */
      for (i = 0; i <= endword; i++)
	row[i] = 0;
      /* calculate scaled v coord of first visible row */
      scaledv = ScaleVpos (top - windowtop) + windowv;

      /* only consider visible rows; thisrow := top to bottom */
      thisrow = top;
      while (true)
      {
	/* move to first byte of first visible word in this row */
	ptr.UU.int_ = pixtab->bitmap.UU.int_ +
		      ((thisrow - vpmyo) * wordsperrow + firstword) * 4;
	/* get row of visible words from bitmap and OR with row array */
	wordpos = 0;
	while (true)
	{
	  row[wordpos] |= *ptr.UU.bptr;   /* set union */
	  if (wordpos == endword)
	    break;  /* escape from inner "infinite" loop */
	  wordpos++;
	  ptr.UU.int_ += 4;   /* next word */
	}
	/* calculate scaled v coord of next row */
	scalednextv = ScaleVpos (thisrow - windowtop + 1) + windowv;
	scaledheight = scalednextv - scaledv;
	if (scaledheight > 0 || thisrow == bottom)
	{
	  /* display black pixels in row, doing any h/v expansion */
	  if (scaledheight < 1)   /* avoid 0 */
	    scaledheight = 1;
	  inrun = false;
	  /* bitpos ranges over (WORDSIZE-1)..0 */
	  bitpos = (WORDSIZE-1) - (firstbit & (WORDSIZE-1));
	  wordpos = 0;

	  /* only consider visible bits; thisbit := left to right */
	  thisbit = left;
	  while (true)  /* bit loop */
	  {
	    if ((unsigned) bitpos < WORDSIZE
	        && ((1 << bitpos) & row[wordpos]) != 0)
	    {  /* start/continue run */
	      if (!inrun)
	      {  /* remember start of run */
		inrun = true;
		scaledh = ScaleHpos (thisbit - windowleft) + windowh;
	      }
	    }
	    else if (inrun)
	    {
	      inrun = false;
	      scaledwidth = ScaleHpos (thisbit - windowleft) + windowh - scaledh;
	      if (scaledwidth < 1)   /* avoid 0 */
		scaledwidth = 1;
	      ShowRectangle (scaledh, scaledv, scaledwidth, scaledheight,
			    chartab->code);
	    }
	    if (thisbit == right)   /* EXIT bit loop */
	      break;
	    if (bitpos == 0)
	    {
	      wordpos++;
	      bitpos = (WORDSIZE-1);
	    }
	    else  /* look at next bit in word */
	    {
	      bitpos--;
	    }
	    thisbit++;
	  }  /* bit loop */

	  if (inrun)
	  {  /* show run at end of row */
	    scaledwidth = ScaleHpos (thisbit - windowleft + 1) + windowh - scaledh;
	    if (scaledwidth < 1)   /* avoid 0 */
	      scaledwidth = 1;
	    ShowRectangle (scaledh, scaledv, scaledwidth, scaledheight,
			  chartab->code);
	  }

	  if (thisrow == bottom)   /* EXIT row loop */
	    break;
	  /* else reset the visible words in row to 0 */
	  for (i = 0; i <= endword; i++)
	    row[i] = 0;
	} /* if */
	scaledv = scalednextv;
	thisrow++;
      }  /* row loop */

    } /* if */
    /* visible part */
  }


  /* 0 bit has ended run */
  return true;
}
/* FullChar */


#undef WORDSIZE
#undef MAXVISWORDS

/******************************************************************************/

static void DisplayChars (void)
{
  /* Display all characters on a font by font basis.  How characters will be
     represented depends on the current displaymode (which the user can change,
     while the window is being updated, by typing the Terse/Box/Full commands).
     Fonts will be displayed in order of ascending totalchars (due to SortFonts).
     Characters in a font will be displayed in a top-down, left-right manner
     because of the way DVIReader builds a charlist.
  */

  for (thisfontinfo = fontlist;
       thisfontinfo != unusedfont;
       thisfontinfo = thisfontinfo->nextfont)
  {
    /* SortFont makes sure we only consider used fonts */

    fontopen = false;   /* might be set in FullChar */

    /* Some VDUs may be able to simulate the given font.
       To help the VDU select appropriately sized characters, we need to
       pass the scaledsize of the font (converted to unscaled paper pixels),
       the overall mag, and the current h/vscalefactors.
    */

    LoadFont (thisfontinfo->fontspec,
              XPixelRound (thisfontinfo->scaledsize),
              mag / 1000.0,
              hscalefactor,
              vscalefactor);

    /* display chars in chartable */

    for (thischarinfo = thisfontinfo->charlist;
         thischarinfo != (charinfo *) NULL;
	 thischarinfo = thischarinfo->nextchar)
    {
      for (thischar = 0;  thischar < thischarinfo->charcount;  thischar++)
      {
        if (thisfontinfo->fontexists)
	{
	  if (displaymode & fullmode)
	  {
	    if (thisfontinfo->pkfont)
	      FullChar();
	    else
#ifdef FULL_CHAR_PS
	      FullCharPS();
#else
	      BoxChar();
#endif
	  }
	  if (displaymode & tersemode)
	  {
	    TerseChar();
	  }
	  if (displaymode & boxmode)
	  {
	    BoxChar();
	  }
	}
        else  /* font is missing! */
        {
	  if (displaymode & (fullmode | tersemode))
	  {
	    TerseChar();
	  }
	  if (displaymode & boxmode)
	  {
	    BoxChar();
	  }
        }
      }  /* inner (for) loop */
    }  /* middle (for) loop */
    if (fontopen)
      CloseFontFile();
  }  /* outer (for) loop */
}
/* DisplayChars */

/******************************************************************************/

#ifdef __STDC__
static void PaperMessage (double precision)
#else
static void PaperMessage (precision)
    double precision;
#endif
{
  /* Called by CheckPageEdges to remind user of the paper size. */

  WriteUnits();

  MesgString ("! (Paper ");
  WriteXDim (precision, paperwd);
  MesgString (" x ");
  WriteYDim (precision, paperht);
  MesgChar (')');

  WaitForReturn();
  ClearMessageLine();
}
/* PaperMessage */

/******************************************************************************/

static void CheckPageEdges (void)
{
  /* One or more page edges do not fall within the paper edges.
     This routine is called after the page & paper have been displayed so
     user can see how bad the problem is.
  */

  double precision = 1.0e-2;

  if (minhp < paperleft)
  {
    ClearMessageLine();
    MesgString ("Page past paper's left edge by ");
    WriteXDim (precision, paperleft - minhp);
    PaperMessage (precision);
  }
  if (maxhp > paperright)
  {
    ClearMessageLine();
    MesgString ("Page past paper's right edge by ");
    WriteXDim (precision, maxhp - paperright);
    PaperMessage (precision);
  }
  if (minvp < papertop)
  {
    ClearMessageLine();
    MesgString ("Page past paper's top edge by ");
    WriteYDim (precision, papertop - minvp);
    PaperMessage (precision);
  }
  if (maxvp > paperbottom)
  {
    ClearMessageLine();
    MesgString ("Page past paper's bottom edge by ");
    WriteYDim (precision, maxvp - paperbottom);
    PaperMessage (precision);
  }
}
/* CheckPageEdges */

/******************************************************************************/

static void DisplayPage (void)
{
  /* Display page in window region based on window location and size,
     and displaymode.  This routine is only called if paintwindow is TRUE
     after all commands have been processed.
  */

  if (screenjustcleared)
  {
    /* avoid doing it again */

    if (paintDVIStatus)
      UpdateDVIStatusLine();

    if (paintWindowStatus)
      UpdateWindowStatusLine();
  }
  else
  {
      ClearRegion(ScaleHpos(invalidleft - windowleft) + windowh, 
		  ScaleVpos(invalidtop - windowtop) + windowv, 
		  ScaleHpos(invalidright - windowleft) + windowh + 1,  
		  ScaleVpos(invalidbottom - windowtop) + windowv + 1);
    UpdateDVIStatusLine();
    UpdateWindowStatusLine();
  }

  StartGraphics();
  DisplayPaperEdges();
  StartText();

  if (pageempty)
  {
    ClearMessageLine();
    MesgString ("Page is empty.");
  }
  else if (outsidepage)
  {
    if (pageoffpaper)
      CheckPageEdges();

    ClearMessageLine();
    MesgString ("Window is ");

    if (windowtop > maxvp)
    {
      MesgString ("below ");
      if (windowleft > maxhp || windowleft <= minhp - scaledwd)
        MesgString ("and ");
    }
    else if (windowtop <= minvp - scaledht)
    {
      MesgString ("above ");
      if (windowleft > maxhp || windowleft <= minhp - scaledwd)
        MesgString ("and ");
    }

    if (windowleft > maxhp)
      MesgString ("to the right of ");
    else if (windowleft <= minhp - scaledwd)
      MesgString ("to the left of ");

    MesgString ("page.");
  }
  else
  {
    /* Page is not empty and part or all of it is visible. */
    StartGraphics();

    useraborted = false;

    DisplayRules();

    if (!useraborted)
      DisplayChars();

    StartText();

    if (pageoffpaper)
      CheckPageEdges();  /* May write messages */

    if (allpagevisible)
    {
      ClearMessageLine();
      MesgString ("Entire page is visible.  ");
      PadMesg();
    }
  }

  MesgFlush();
}
/* DisplayPage */

/******************************************************************************/
/******************************************************************************/

void ProcessCommand (void)
{
  ClearMessageLine();   /* erase message line at this stage */
  /* initialize flags for multiple command processing */
  badcommand = false;
  paintWindowStatus = false;
  paintDVIStatus = false;
  paintwindow = false;
  screenjustcleared = false;
  pageoffpaper = false;
  Valid ( );
    switch (command)
    {
    case UpCmd:
    case DownCmd:
      WindowUpDown();
      paintWindowStatus = true;
      break;

    case LeftCmd:
    case RightCmd:
      WindowLeftRight();
      paintWindowStatus = true;
      break;

    case ZoomInCmd:  /* Zoom in or out */
    case ZoomOutCmd:
      ZoomWindow();
      break;

    case NextCmd:  /* Next page */
      if (NextPageFound (true))  /* ascending */
        ProcessPage();
      break;

    case PrevCmd:  /* Previous page */
      if (NextPageFound (false))  /* descending */
        ProcessPage();
      break;

    case 'G':
    case 'g':
          DVIPageFound(argument);
          ProcessPage();
          break;

    case 'B':  /* Box display */
      displaymode ^= boxmode;  /* toggle box mode */
      paintDVIStatus = true;
      if (currDVIpage != 0) {
	Invalid(windowtop, windowleft, windowbottom, windowright);
	paintwindow = true;
      }
      break;

    case 'R':
    case 'r':
      DisplayPage();  /* unconditionally, and immediately on interpretation */
      break;

    case 'q': command='Q';
    case 'Q':
      goto _L999;
      break;

    default:  /* unknown command */
      break;
    }  /* command */

  if (paintwindow)   /* only update window after processing all commands */
    DisplayPage();
  else
  {
    if (paintDVIStatus)
      UpdateDVIStatusLine();
    if (paintWindowStatus)
      UpdateWindowStatusLine();
  }
_L999: ;
}
/* ProcessCommandLine */

/******************************************************************************/

static void NextCommandLine (void)
{
  ClearTextLine (commandl);
  MoveToTextLine (commandl);
  MesgString (commprompt);
  MesgFlush();
  command = ReadCommand();
  ProcessCommand ();
}
/* NextCommandLine */

/******************************************************************************/

static void Finish (void)
{
  /* close DVI file, reset VDU and restore terminal */

  CloseDVIFile();

  ClearScreen();
  screenjustcleared = true;

  MoveToTextLine (1);
  MesgLine();

  ResetVDU();
  RestoreTerminal();

}
/* Finish */

/***************************************************************************/

int OutMem (void)
{
  /* used in "config.h" to define Malloc() */

  StartText();
  ResetVDU();   /* do before message since it might erase screen! */

  MesgString ("Out of memory.");
  MesgLine();

  RestoreTerminal();
  exit (1);

  return 0;                        /* Not Reached */
}
/* OutMem */


void DefaultLog (void) { }

/******************************************************************************/

void Handler (int sig)
{
  /* Handle each of the six kinds of signals listed in the ANSI C K&R book. */

  typedef  void (* PH) __((int));
  PH  prev_handler;

  prev_handler = (PH) signal (sig, Handler);
  if (prev_handler == SIG_ERR)
  {
    MesgString ("signal handling failed in Handler");
    MesgLine();
  }

  StartText();
  ResetVDU();   /* do before message since it might erase screen! */
  RestoreTerminal();

  MesgString
    ((sig == SIGABRT ? "SIGABRT = abort signal" :
      sig == SIGFPE  ? "SIGFPE = floating point error" :
      sig == SIGILL  ? "SIGILL = illegal function image, eg, illegal instruction" :
      sig == SIGINT  ? "SIGINT = interactive attention, eg interrupt" :
      sig == SIGSEGV ? "SIGSEGV = illegal storage access, eg outside memory limits" :
      sig == SIGTERM ? "SIGTERM = termination request" :
                       "unknown signal"));
  MesgLine();

  RestoreTerminal();  /* just to be sure */
  exit (1);
}
/* Handler */

/******************************************************************************/

void TrapInterrupts (void)
{
  /* trap interrupts, to allow screen reset before exit */
  typedef  void (* PH) __((int));
  PH  prev_handler;
  int i;
  for (i=0;  i < sizeof (common_signals) / sizeof (int);  i++)
  {
    prev_handler = (PH) signal (common_signals[i], Handler);
    if (prev_handler == SIG_ERR)
    {
      MesgString ("dview:  signal handling failed at outset!");
      MesgLine();
    }
  }
}
/* TrapInterrupts */

/******************************************************************************/

int main (int argc, Char *argv[])
{
  P_argv = argv;
  P_argc = argc;

  textlinewidth = 80;  /* for the initial screen type */ 

  fprintf (stderr, "%s\n", dvgt_version);

  if (argc <= 1)  /* No commandline arguments: assume user is new */
  {
    char * env;
    stringcopy (immed_help, (!(env = getenv ("DV_IMMED_HELP")) ?
			  DEF_IMMED_HELP : env));
    /* If this file cannot be printed, print the precompiled information */
    if (!PrintFile (immed_help))
      PrintText (immed_help_strings);
  }
  else
  {

    InitOptions();   /* initialise DVIname and command options */
    if (DVIname == (char *) NULL ||  strlen (DVIname) == 0
				 ||  DVIname[0] == ' ')
    {
      /* gt - Eventually, I want to load DVI files interactively. */
      DefaultLog();  /* NOT REACHED; JUST EXITED ! */
      MesgString ("This line is not reached !!");
      MesgLine();
    }
    else
    {
      Char STR1[2];

      TrapInterrupts();
      InitScreenIO();   /* SYSDEP: sets cbreak on & echo off */

      InitDVIReader();
      InitFontReader();
      OpenDVIFile (DVIname);   /* and read DVImag etc. */
      if (mag == 0)   /* use DVImag */
	mag = DVImag;
      SetConversionFactors (xres, yres, mag/1000.0);   /* for DVIReader */
      InitGGI();   /* init windowwd/ht etc. */
      InitWinVars();
      StartText();
      ClearScreen();
      screenjustcleared = true;
      UpdateDVIStatusLine();
      UpdateWindowStatusLine();
      MoveToTextLine (messagel);
      MesgString (dvgt_version);
      MesgLine();
      do
      {
	NextCommandLine();   /* parse and execute command(s) */
      } while (strcmp ((sprintf (STR1, "%c", command), STR1), Quit));
      Finish();
    }  /* if */
  }  /* if */
  return 0;  /* OK */
}
/* main() program of DVItoVDU */

/* end dvitovdu.c */
